# Generated from GraphQL.g4 by ANTLR 4.9
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .GraphQLParser import GraphQLParser
else:
    from GraphQLParser import GraphQLParser

# This class defines a complete listener for a parse tree produced by GraphQLParser.
class GraphQLListener(ParseTreeListener):

    # Enter a parse tree produced by GraphQLParser#document.
    def enterDocument(self, ctx:GraphQLParser.DocumentContext):
        pass

    # Exit a parse tree produced by GraphQLParser#document.
    def exitDocument(self, ctx:GraphQLParser.DocumentContext):
        pass


    # Enter a parse tree produced by GraphQLParser#definition.
    def enterDefinition(self, ctx:GraphQLParser.DefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#definition.
    def exitDefinition(self, ctx:GraphQLParser.DefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#executableDefinition.
    def enterExecutableDefinition(self, ctx:GraphQLParser.ExecutableDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#executableDefinition.
    def exitExecutableDefinition(self, ctx:GraphQLParser.ExecutableDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#operationDefinition.
    def enterOperationDefinition(self, ctx:GraphQLParser.OperationDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#operationDefinition.
    def exitOperationDefinition(self, ctx:GraphQLParser.OperationDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#operationType.
    def enterOperationType(self, ctx:GraphQLParser.OperationTypeContext):
        pass

    # Exit a parse tree produced by GraphQLParser#operationType.
    def exitOperationType(self, ctx:GraphQLParser.OperationTypeContext):
        pass


    # Enter a parse tree produced by GraphQLParser#selectionSet.
    def enterSelectionSet(self, ctx:GraphQLParser.SelectionSetContext):
        pass

    # Exit a parse tree produced by GraphQLParser#selectionSet.
    def exitSelectionSet(self, ctx:GraphQLParser.SelectionSetContext):
        pass


    # Enter a parse tree produced by GraphQLParser#selection.
    def enterSelection(self, ctx:GraphQLParser.SelectionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#selection.
    def exitSelection(self, ctx:GraphQLParser.SelectionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#field.
    def enterField(self, ctx:GraphQLParser.FieldContext):
        pass

    # Exit a parse tree produced by GraphQLParser#field.
    def exitField(self, ctx:GraphQLParser.FieldContext):
        pass


    # Enter a parse tree produced by GraphQLParser#arguments.
    def enterArguments(self, ctx:GraphQLParser.ArgumentsContext):
        pass

    # Exit a parse tree produced by GraphQLParser#arguments.
    def exitArguments(self, ctx:GraphQLParser.ArgumentsContext):
        pass


    # Enter a parse tree produced by GraphQLParser#argument.
    def enterArgument(self, ctx:GraphQLParser.ArgumentContext):
        pass

    # Exit a parse tree produced by GraphQLParser#argument.
    def exitArgument(self, ctx:GraphQLParser.ArgumentContext):
        pass


    # Enter a parse tree produced by GraphQLParser#alias.
    def enterAlias(self, ctx:GraphQLParser.AliasContext):
        pass

    # Exit a parse tree produced by GraphQLParser#alias.
    def exitAlias(self, ctx:GraphQLParser.AliasContext):
        pass


    # Enter a parse tree produced by GraphQLParser#fragmentSpread.
    def enterFragmentSpread(self, ctx:GraphQLParser.FragmentSpreadContext):
        pass

    # Exit a parse tree produced by GraphQLParser#fragmentSpread.
    def exitFragmentSpread(self, ctx:GraphQLParser.FragmentSpreadContext):
        pass


    # Enter a parse tree produced by GraphQLParser#fragmentDefinition.
    def enterFragmentDefinition(self, ctx:GraphQLParser.FragmentDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#fragmentDefinition.
    def exitFragmentDefinition(self, ctx:GraphQLParser.FragmentDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#fragmentName.
    def enterFragmentName(self, ctx:GraphQLParser.FragmentNameContext):
        pass

    # Exit a parse tree produced by GraphQLParser#fragmentName.
    def exitFragmentName(self, ctx:GraphQLParser.FragmentNameContext):
        pass


    # Enter a parse tree produced by GraphQLParser#typeCondition.
    def enterTypeCondition(self, ctx:GraphQLParser.TypeConditionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#typeCondition.
    def exitTypeCondition(self, ctx:GraphQLParser.TypeConditionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#inlineFragment.
    def enterInlineFragment(self, ctx:GraphQLParser.InlineFragmentContext):
        pass

    # Exit a parse tree produced by GraphQLParser#inlineFragment.
    def exitInlineFragment(self, ctx:GraphQLParser.InlineFragmentContext):
        pass


    # Enter a parse tree produced by GraphQLParser#value.
    def enterValue(self, ctx:GraphQLParser.ValueContext):
        pass

    # Exit a parse tree produced by GraphQLParser#value.
    def exitValue(self, ctx:GraphQLParser.ValueContext):
        pass


    # Enter a parse tree produced by GraphQLParser#intValue.
    def enterIntValue(self, ctx:GraphQLParser.IntValueContext):
        pass

    # Exit a parse tree produced by GraphQLParser#intValue.
    def exitIntValue(self, ctx:GraphQLParser.IntValueContext):
        pass


    # Enter a parse tree produced by GraphQLParser#floatValue.
    def enterFloatValue(self, ctx:GraphQLParser.FloatValueContext):
        pass

    # Exit a parse tree produced by GraphQLParser#floatValue.
    def exitFloatValue(self, ctx:GraphQLParser.FloatValueContext):
        pass


    # Enter a parse tree produced by GraphQLParser#booleanValue.
    def enterBooleanValue(self, ctx:GraphQLParser.BooleanValueContext):
        pass

    # Exit a parse tree produced by GraphQLParser#booleanValue.
    def exitBooleanValue(self, ctx:GraphQLParser.BooleanValueContext):
        pass


    # Enter a parse tree produced by GraphQLParser#stringValue.
    def enterStringValue(self, ctx:GraphQLParser.StringValueContext):
        pass

    # Exit a parse tree produced by GraphQLParser#stringValue.
    def exitStringValue(self, ctx:GraphQLParser.StringValueContext):
        pass


    # Enter a parse tree produced by GraphQLParser#nullValue.
    def enterNullValue(self, ctx:GraphQLParser.NullValueContext):
        pass

    # Exit a parse tree produced by GraphQLParser#nullValue.
    def exitNullValue(self, ctx:GraphQLParser.NullValueContext):
        pass


    # Enter a parse tree produced by GraphQLParser#enumValue.
    def enterEnumValue(self, ctx:GraphQLParser.EnumValueContext):
        pass

    # Exit a parse tree produced by GraphQLParser#enumValue.
    def exitEnumValue(self, ctx:GraphQLParser.EnumValueContext):
        pass


    # Enter a parse tree produced by GraphQLParser#listValue.
    def enterListValue(self, ctx:GraphQLParser.ListValueContext):
        pass

    # Exit a parse tree produced by GraphQLParser#listValue.
    def exitListValue(self, ctx:GraphQLParser.ListValueContext):
        pass


    # Enter a parse tree produced by GraphQLParser#objectValue.
    def enterObjectValue(self, ctx:GraphQLParser.ObjectValueContext):
        pass

    # Exit a parse tree produced by GraphQLParser#objectValue.
    def exitObjectValue(self, ctx:GraphQLParser.ObjectValueContext):
        pass


    # Enter a parse tree produced by GraphQLParser#objectField.
    def enterObjectField(self, ctx:GraphQLParser.ObjectFieldContext):
        pass

    # Exit a parse tree produced by GraphQLParser#objectField.
    def exitObjectField(self, ctx:GraphQLParser.ObjectFieldContext):
        pass


    # Enter a parse tree produced by GraphQLParser#variable.
    def enterVariable(self, ctx:GraphQLParser.VariableContext):
        pass

    # Exit a parse tree produced by GraphQLParser#variable.
    def exitVariable(self, ctx:GraphQLParser.VariableContext):
        pass


    # Enter a parse tree produced by GraphQLParser#variableDefinitions.
    def enterVariableDefinitions(self, ctx:GraphQLParser.VariableDefinitionsContext):
        pass

    # Exit a parse tree produced by GraphQLParser#variableDefinitions.
    def exitVariableDefinitions(self, ctx:GraphQLParser.VariableDefinitionsContext):
        pass


    # Enter a parse tree produced by GraphQLParser#variableDefinition.
    def enterVariableDefinition(self, ctx:GraphQLParser.VariableDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#variableDefinition.
    def exitVariableDefinition(self, ctx:GraphQLParser.VariableDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#defaultValue.
    def enterDefaultValue(self, ctx:GraphQLParser.DefaultValueContext):
        pass

    # Exit a parse tree produced by GraphQLParser#defaultValue.
    def exitDefaultValue(self, ctx:GraphQLParser.DefaultValueContext):
        pass


    # Enter a parse tree produced by GraphQLParser#type_.
    def enterType_(self, ctx:GraphQLParser.Type_Context):
        pass

    # Exit a parse tree produced by GraphQLParser#type_.
    def exitType_(self, ctx:GraphQLParser.Type_Context):
        pass


    # Enter a parse tree produced by GraphQLParser#namedType.
    def enterNamedType(self, ctx:GraphQLParser.NamedTypeContext):
        pass

    # Exit a parse tree produced by GraphQLParser#namedType.
    def exitNamedType(self, ctx:GraphQLParser.NamedTypeContext):
        pass


    # Enter a parse tree produced by GraphQLParser#listType.
    def enterListType(self, ctx:GraphQLParser.ListTypeContext):
        pass

    # Exit a parse tree produced by GraphQLParser#listType.
    def exitListType(self, ctx:GraphQLParser.ListTypeContext):
        pass


    # Enter a parse tree produced by GraphQLParser#directives.
    def enterDirectives(self, ctx:GraphQLParser.DirectivesContext):
        pass

    # Exit a parse tree produced by GraphQLParser#directives.
    def exitDirectives(self, ctx:GraphQLParser.DirectivesContext):
        pass


    # Enter a parse tree produced by GraphQLParser#directive.
    def enterDirective(self, ctx:GraphQLParser.DirectiveContext):
        pass

    # Exit a parse tree produced by GraphQLParser#directive.
    def exitDirective(self, ctx:GraphQLParser.DirectiveContext):
        pass


    # Enter a parse tree produced by GraphQLParser#typeSystemDefinition.
    def enterTypeSystemDefinition(self, ctx:GraphQLParser.TypeSystemDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#typeSystemDefinition.
    def exitTypeSystemDefinition(self, ctx:GraphQLParser.TypeSystemDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#typeSystemExtension.
    def enterTypeSystemExtension(self, ctx:GraphQLParser.TypeSystemExtensionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#typeSystemExtension.
    def exitTypeSystemExtension(self, ctx:GraphQLParser.TypeSystemExtensionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#schemaDefinition.
    def enterSchemaDefinition(self, ctx:GraphQLParser.SchemaDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#schemaDefinition.
    def exitSchemaDefinition(self, ctx:GraphQLParser.SchemaDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#rootOperationTypeDefinition.
    def enterRootOperationTypeDefinition(self, ctx:GraphQLParser.RootOperationTypeDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#rootOperationTypeDefinition.
    def exitRootOperationTypeDefinition(self, ctx:GraphQLParser.RootOperationTypeDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#schemaExtension.
    def enterSchemaExtension(self, ctx:GraphQLParser.SchemaExtensionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#schemaExtension.
    def exitSchemaExtension(self, ctx:GraphQLParser.SchemaExtensionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#operationTypeDefinition.
    def enterOperationTypeDefinition(self, ctx:GraphQLParser.OperationTypeDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#operationTypeDefinition.
    def exitOperationTypeDefinition(self, ctx:GraphQLParser.OperationTypeDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#description.
    def enterDescription(self, ctx:GraphQLParser.DescriptionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#description.
    def exitDescription(self, ctx:GraphQLParser.DescriptionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#typeDefinition.
    def enterTypeDefinition(self, ctx:GraphQLParser.TypeDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#typeDefinition.
    def exitTypeDefinition(self, ctx:GraphQLParser.TypeDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#typeExtension.
    def enterTypeExtension(self, ctx:GraphQLParser.TypeExtensionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#typeExtension.
    def exitTypeExtension(self, ctx:GraphQLParser.TypeExtensionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#scalarTypeDefinition.
    def enterScalarTypeDefinition(self, ctx:GraphQLParser.ScalarTypeDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#scalarTypeDefinition.
    def exitScalarTypeDefinition(self, ctx:GraphQLParser.ScalarTypeDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#scalarTypeExtension.
    def enterScalarTypeExtension(self, ctx:GraphQLParser.ScalarTypeExtensionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#scalarTypeExtension.
    def exitScalarTypeExtension(self, ctx:GraphQLParser.ScalarTypeExtensionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#objectTypeDefinition.
    def enterObjectTypeDefinition(self, ctx:GraphQLParser.ObjectTypeDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#objectTypeDefinition.
    def exitObjectTypeDefinition(self, ctx:GraphQLParser.ObjectTypeDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#implementsInterfaces.
    def enterImplementsInterfaces(self, ctx:GraphQLParser.ImplementsInterfacesContext):
        pass

    # Exit a parse tree produced by GraphQLParser#implementsInterfaces.
    def exitImplementsInterfaces(self, ctx:GraphQLParser.ImplementsInterfacesContext):
        pass


    # Enter a parse tree produced by GraphQLParser#fieldsDefinition.
    def enterFieldsDefinition(self, ctx:GraphQLParser.FieldsDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#fieldsDefinition.
    def exitFieldsDefinition(self, ctx:GraphQLParser.FieldsDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#fieldDefinition.
    def enterFieldDefinition(self, ctx:GraphQLParser.FieldDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#fieldDefinition.
    def exitFieldDefinition(self, ctx:GraphQLParser.FieldDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#argumentsDefinition.
    def enterArgumentsDefinition(self, ctx:GraphQLParser.ArgumentsDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#argumentsDefinition.
    def exitArgumentsDefinition(self, ctx:GraphQLParser.ArgumentsDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#inputValueDefinition.
    def enterInputValueDefinition(self, ctx:GraphQLParser.InputValueDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#inputValueDefinition.
    def exitInputValueDefinition(self, ctx:GraphQLParser.InputValueDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#objectTypeExtension.
    def enterObjectTypeExtension(self, ctx:GraphQLParser.ObjectTypeExtensionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#objectTypeExtension.
    def exitObjectTypeExtension(self, ctx:GraphQLParser.ObjectTypeExtensionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#interfaceTypeDefinition.
    def enterInterfaceTypeDefinition(self, ctx:GraphQLParser.InterfaceTypeDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#interfaceTypeDefinition.
    def exitInterfaceTypeDefinition(self, ctx:GraphQLParser.InterfaceTypeDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#interfaceTypeExtension.
    def enterInterfaceTypeExtension(self, ctx:GraphQLParser.InterfaceTypeExtensionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#interfaceTypeExtension.
    def exitInterfaceTypeExtension(self, ctx:GraphQLParser.InterfaceTypeExtensionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#unionTypeDefinition.
    def enterUnionTypeDefinition(self, ctx:GraphQLParser.UnionTypeDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#unionTypeDefinition.
    def exitUnionTypeDefinition(self, ctx:GraphQLParser.UnionTypeDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#unionMemberTypes.
    def enterUnionMemberTypes(self, ctx:GraphQLParser.UnionMemberTypesContext):
        pass

    # Exit a parse tree produced by GraphQLParser#unionMemberTypes.
    def exitUnionMemberTypes(self, ctx:GraphQLParser.UnionMemberTypesContext):
        pass


    # Enter a parse tree produced by GraphQLParser#unionTypeExtension.
    def enterUnionTypeExtension(self, ctx:GraphQLParser.UnionTypeExtensionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#unionTypeExtension.
    def exitUnionTypeExtension(self, ctx:GraphQLParser.UnionTypeExtensionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#enumTypeDefinition.
    def enterEnumTypeDefinition(self, ctx:GraphQLParser.EnumTypeDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#enumTypeDefinition.
    def exitEnumTypeDefinition(self, ctx:GraphQLParser.EnumTypeDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#enumValuesDefinition.
    def enterEnumValuesDefinition(self, ctx:GraphQLParser.EnumValuesDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#enumValuesDefinition.
    def exitEnumValuesDefinition(self, ctx:GraphQLParser.EnumValuesDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#enumValueDefinition.
    def enterEnumValueDefinition(self, ctx:GraphQLParser.EnumValueDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#enumValueDefinition.
    def exitEnumValueDefinition(self, ctx:GraphQLParser.EnumValueDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#enumTypeExtension.
    def enterEnumTypeExtension(self, ctx:GraphQLParser.EnumTypeExtensionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#enumTypeExtension.
    def exitEnumTypeExtension(self, ctx:GraphQLParser.EnumTypeExtensionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#inputObjectTypeDefinition.
    def enterInputObjectTypeDefinition(self, ctx:GraphQLParser.InputObjectTypeDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#inputObjectTypeDefinition.
    def exitInputObjectTypeDefinition(self, ctx:GraphQLParser.InputObjectTypeDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#inputFieldsDefinition.
    def enterInputFieldsDefinition(self, ctx:GraphQLParser.InputFieldsDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#inputFieldsDefinition.
    def exitInputFieldsDefinition(self, ctx:GraphQLParser.InputFieldsDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#inputObjectTypeExtension.
    def enterInputObjectTypeExtension(self, ctx:GraphQLParser.InputObjectTypeExtensionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#inputObjectTypeExtension.
    def exitInputObjectTypeExtension(self, ctx:GraphQLParser.InputObjectTypeExtensionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#directiveDefinition.
    def enterDirectiveDefinition(self, ctx:GraphQLParser.DirectiveDefinitionContext):
        pass

    # Exit a parse tree produced by GraphQLParser#directiveDefinition.
    def exitDirectiveDefinition(self, ctx:GraphQLParser.DirectiveDefinitionContext):
        pass


    # Enter a parse tree produced by GraphQLParser#directiveLocations.
    def enterDirectiveLocations(self, ctx:GraphQLParser.DirectiveLocationsContext):
        pass

    # Exit a parse tree produced by GraphQLParser#directiveLocations.
    def exitDirectiveLocations(self, ctx:GraphQLParser.DirectiveLocationsContext):
        pass


    # Enter a parse tree produced by GraphQLParser#directiveLocation.
    def enterDirectiveLocation(self, ctx:GraphQLParser.DirectiveLocationContext):
        pass

    # Exit a parse tree produced by GraphQLParser#directiveLocation.
    def exitDirectiveLocation(self, ctx:GraphQLParser.DirectiveLocationContext):
        pass


    # Enter a parse tree produced by GraphQLParser#executableDirectiveLocation.
    def enterExecutableDirectiveLocation(self, ctx:GraphQLParser.ExecutableDirectiveLocationContext):
        pass

    # Exit a parse tree produced by GraphQLParser#executableDirectiveLocation.
    def exitExecutableDirectiveLocation(self, ctx:GraphQLParser.ExecutableDirectiveLocationContext):
        pass


    # Enter a parse tree produced by GraphQLParser#typeSystemDirectiveLocation.
    def enterTypeSystemDirectiveLocation(self, ctx:GraphQLParser.TypeSystemDirectiveLocationContext):
        pass

    # Exit a parse tree produced by GraphQLParser#typeSystemDirectiveLocation.
    def exitTypeSystemDirectiveLocation(self, ctx:GraphQLParser.TypeSystemDirectiveLocationContext):
        pass


    # Enter a parse tree produced by GraphQLParser#name.
    def enterName(self, ctx:GraphQLParser.NameContext):
        pass

    # Exit a parse tree produced by GraphQLParser#name.
    def exitName(self, ctx:GraphQLParser.NameContext):
        pass



del GraphQLParser


from antlr4 import *
from .antlr import GraphQLLexer, GraphQLListener, GraphQLParser
from .codegen import CodegenTool, Class, String, ClassInstance, IfElse, If, Method, Expr, Variable
import re
from math import floor
from datetime import datetime
from .utils import strip_string_quotes, camel_case_to_snake_case, process_input_value_definition
from .errors import ParsingError

GraphQLParser = GraphQLParser.GraphQLParser

graphene = 'graphene'
built_in_scalars = [
    'Int',
    'Float',
    'String',
    'Boolean',
    'ID',
    'Date',
    'Datetime',
    'Time'
    'Decimal',
    'JSONString',
    'Base64',
]


class SchemaGen(GraphQLListener.GraphQLListener):
    """
    SchemaGen is the entry point through which the package is used.

    Attributes:
        input_file: a string containing the name of the GraphQL schema file
        output_file: an optional string containing the name of the file to which the result of the code generation should be written to.
    """

    def __init__(self, input_file: str, output_file: str = None):
        if output_file is None:
            output_file = input_file.split(sep='.')[0] + '_' + str(floor(datetime.now().timestamp())) + '.py'

        is_valid_file_name = re.match("\w+.py$", output_file)
        if is_valid_file_name is None:
            raise Exception('File is not a python file')
        self.output_file = output_file
        self.input_file = input_file
        self.codegen = CodegenTool(output_file=self.output_file)
        super().__init__()

    def enterTypeDefinition(self, ctx: GraphQLParser.TypeDefinitionContext):

        for child in ctx.children:
            # type definition is for an Object Type Definition
            if isinstance(child, GraphQLParser.ObjectTypeDefinitionContext) or isinstance(child,
                                                                                          GraphQLParser.InterfaceTypeDefinitionContext):
                # print(child.name().getText())
                # if ":" in child.name().getText():
                #     child.exitRule(self)
                is_object_type = isinstance(child, GraphQLParser.ObjectTypeDefinitionContext)
                is_interface = isinstance(child, GraphQLParser.InterfaceTypeDefinitionContext)

                type_class = Class(name=child.name().getText(), add_init_method=False)
                if is_object_type:
                    type_class.base_class = "ObjectType"
                elif is_interface:
                    type_class.base_class = "Interface"

                is_mutation = False
                if type_class.name == 'Mutation':
                    is_mutation = True
                    is_object_type = False

                meta_class = Class(name='meta')

                # create map for methods to be resolved
                methods_to_be_resolved = {}

                # get type description
                desc = child.description()
                if desc:
                    meta_class.add_class_variable('description', String(strip_string_quotes(desc.getText())))

                # get implemented interfaces
                if is_object_type or is_mutation:
                    if child.implementsInterfaces() is not None:
                        interfaces = child.implementsInterfaces().getText().split(sep='implements')
                        interfaces = interfaces[1].split(sep='&')
                        interface_string = ''
                        for i in interfaces:
                            interface_string = interface_string + i + ','
                        meta_class.add_class_variable('interfaces', f"({interface_string})")

                # get fields of the ObjectType or Interface

                if child.fieldsDefinition():
                    fields = child.fieldsDefinition().fields
                    if not is_mutation:
                        for field in fields:
                            # get field name and type
                            field_name = camel_case_to_snake_case(field.name().getText())
                            field_type = field.type_().getText()
                            field_required = False

                            # get field description
                            field_desc = field.description()
                            if field_desc is not None:
                                field_desc = String(strip_string_quotes(field_desc.getText()))
                            else:
                                field_desc = ''

                            if is_interface:
                                if field_type.lower() == type_class.name.lower():
                                    field_type = 'lambda: ' + field_type

                            # if field is a required field
                            if field_type[len(field_type) - 1] == '!':
                                field_required = True
                                field_code = ClassInstance('Field', field_type[:-1], required=True)
                            else:
                                field_code = ClassInstance('Field', field_type)

                            # if field is a list type
                            if field.type_().listType() is not None:
                                list_type_named_type = field.type_().listType().type_().getText()

                                if is_interface:
                                    if list_type_named_type.lower() == type_class.name.lower():
                                        list_type_named_type = 'lambda: ' + list_type_named_type

                                if list_type_named_type[len(list_type_named_type) - 1] == '!':
                                    field_code = ClassInstance('List',
                                                               str(ClassInstance('NonNull', list_type_named_type[:-1])),
                                                               required=field_required)
                                else:
                                    field_code = ClassInstance('List', list_type_named_type, required=field_required)

                            # get field arguments
                            if is_object_type:
                                args = field.argumentsDefinition()
                                args_string = []
                                if args is not None:
                                    args = args.args
                                    for arg in args:
                                        # add info to method_to_be_resolved map
                                        if field_name not in methods_to_be_resolved:
                                            methods_to_be_resolved[field_name] = [arg.name().getText()]
                                        else:
                                            methods_to_be_resolved[field_name].append(arg.name().getText())
                                        processed_arg = process_input_value_definition(arg)
                                        args_string.append(
                                            f"{String(processed_arg['name'])}: {str(processed_arg['arg_impl'])}")

                                    field_code.add_kwarg('args', "{" + ', '.join(args_string) + "}")

                            if field_desc != '':
                                field_code.add_kwarg(key='description', value=field_desc)
                            type_class.class_variables[field_name] = str(field_code)
                    else:
                        for field in fields:
                            # get field name and type
                            field_name = camel_case_to_snake_case(field.name().getText())
                            field_type = field.type_().getText()
                            field_required = False

                            field_class = Class(field.name().getText(), add_init_method=False, base_class='Mutation')
                            argument_class = Class(name='arguments')

                            # get field description
                            field_desc = field.description()
                            if field_desc is not None:
                                field_desc = String(strip_string_quotes(field_desc.getText()))
                            else:
                                field_desc = ''

                            # if field is a required field
                            if field_type[len(field_type) - 1] == '!':
                                field_required = True
                                field_code = ClassInstance('Field', field_type[:-1], required=True)
                            else:
                                field_code = ClassInstance('Field', field_type)

                            # if field is a list type
                            if field.type_().listType() is not None:
                                list_type_named_type = field.type_().listType().type_().getText()

                                if list_type_named_type[len(list_type_named_type) - 1] == '!':
                                    field_code = ClassInstance('List',
                                                               str(ClassInstance('NonNull', list_type_named_type[:-1])),
                                                               required=field_required)
                                else:
                                    field_code = ClassInstance('List', list_type_named_type, required=field_required)

                            # get field arguments
                            args = field.argumentsDefinition()
                            arg_list = []
                            if args is not None:
                                args = args.args
                                for arg in args:
                                    processed_arg = process_input_value_definition(arg)
                                    argument_class.add_class_variable(processed_arg['name'],
                                                                      str(processed_arg['arg_impl']))
                                    arg_list.append(processed_arg['name'])

                            field_class.add_sub_class(argument_class)
                            field_class.add_method(
                                method=Method(
                                    name='mutate',
                                    arguments=['root', 'info'] + arg_list
                                )
                            )

                            if field_desc != '':
                                field_code.add_kwarg(key='description', value=field_desc)

                            # write mutation classes for the mutation's fields
                            self.codegen.write_class(field_class)
                            type_class.class_variables[field_name] = str(field_code)

                # add resolver methods
                if not is_mutation:
                    for method in methods_to_be_resolved:
                        type_class.add_method(method_name='resolve_' + method,
                                              arguments_names=['info'] + methods_to_be_resolved[method])

                if type_class.name == 'Query':
                    for var in type_class.class_variables:
                        if var not in methods_to_be_resolved:
                            type_class.add_method(method_name='resolve_' + var, arguments_names=['info'])

                if len(meta_class.class_variables) != 0:
                    type_class.add_sub_class(meta_class)

                self.codegen.write_class(type_class)

            # type definition is for an EnumTypeDefinition
            elif isinstance(child, GraphQLParser.EnumTypeDefinitionContext):
                enum_class = Class(name=child.name().getText(), base_class="Enum", add_init_method=False)
                meta_class = Class(name='meta')

                # get enum description
                desc = child.description()
                if desc:
                    meta_class.add_class_variable('description', String(strip_string_quotes(desc.getText())))

                # get fields of the Enum
                fields = child.enumValuesDefinition().fields
                fields_and_desc = {}
                for field in fields:
                    # get field name and type
                    enum_value = field.enumValue().getText()

                    # get enum description
                    field_desc = field.description()
                    if field_desc is not None:
                        field_desc = String(strip_string_quotes(field_desc.getText()))
                    else:
                        field_desc = ''

                    if field_desc != '':
                        # do something
                        fields_and_desc[enum_value] = field_desc

                    # add enums as class variables to main class
                    enum_class.add_class_variable(enum_value, String(enum_value))

                # add enums description
                method = Method(
                    name='description',
                    decorators=['@property'],
                    arguments=[]
                )

                if_else = IfElse(
                    indent_level=method.get_indent_level() + 1,
                    else_action=[Expr("pass")],
                )

                for i in fields_and_desc:
                    if_else.add_elif(If(
                        expr=Expr(f"self == {enum_class.name}.{i}"),
                        action=[Expr(f"return {fields_and_desc[i]}")]
                    ))

                method.set_body([if_else])

                enum_class.add_method(method=method)

                if len(meta_class.class_variables) != 0:
                    enum_class.add_sub_class(meta_class)

                self.codegen.write_class(enum_class)

            # type definition is for an EnumTypeDefinition
            elif isinstance(child, GraphQLParser.ScalarTypeDefinitionContext):
                if child.name().getText().capitalize() in built_in_scalars:
                    continue
                scalar_class = Class(name=child.name().getText(), base_class="Scalar", add_init_method=False)
                desc = child.description()

                if desc is not None:
                    scalar_class.description = strip_string_quotes(desc.getText())

                serialize_method = Method(
                    name='serialize',
                    arguments=['val'],
                    decorators=['@staticmethod'],
                    body=[Expr('# write method body'), Expr('pass')],
                    is_static=True
                )
                parse_literal_method = Method(
                    name='parse_literal',
                    arguments=['node'],
                    decorators=['@staticmethod'],
                    body=[Expr('# write method body'), Expr('pass')],
                    is_static=True
                )

                parse_value_method = Method(
                    name='parse_value',
                    arguments=['value'],
                    decorators=['@staticmethod'],
                    body=[Expr('# write method body'), Expr('pass')],
                    is_static=True
                )

                scalar_class.add_method(method=serialize_method)
                scalar_class.add_method(method=parse_literal_method)
                scalar_class.add_method(method=parse_value_method)

                self.codegen.write_class(scalar_class)

            elif isinstance(child, GraphQLParser.UnionTypeDefinitionContext):
                union_class = Class(name=child.name().getText(), base_class='Union')
                meta_class = Class(name='Meta')

                unions = child.unionMemberTypes().getText()

                if unions[0] == '=':
                    unions = unions[1:]

                unions = ', '.join(unions.split(sep='|'))

                meta_class.add_class_variable(variable_name='types', variable_value=f"({unions})")

                desc = child.description()

                if desc is not None:
                    meta_class.add_class_variable(variable_name='description',
                                                  variable_value=String(strip_string_quotes(desc.getText())))

                union_class.add_sub_class(meta_class)
                self.codegen.write_class(union_class)
                print(unions)

            elif isinstance(child, GraphQLParser.InputObjectTypeDefinitionContext):
                type_class = Class(name=child.name().getText(), base_class="InputObjectType", add_init_method=False)
                meta_class = Class(name='meta')

                # get type description
                desc = child.description()
                if desc:
                    meta_class.add_class_variable('description', String(strip_string_quotes(desc.getText())))

                # get fields
                if child.inputFieldsDefinition():
                    fields = child.inputFieldsDefinition().fields
                    for field in fields:
                        processed_ivd = process_input_value_definition(field)
                        type_class.add_class_variable(processed_ivd['name'], str(processed_ivd['arg_impl']))

                if len(meta_class.class_variables) != 0:
                    type_class.add_sub_class(meta_class)

                self.codegen.write_class(type_class)
            else:
                print(type(child))

    def enterSchemaDefinition(self, ctx: GraphQLParser.SchemaDefinitionContext):

        schema_obj = ClassInstance('Schema')

        fields = ctx.fields
        for field in fields:
            schema_obj.add_kwarg(strip_string_quotes(field.operationType().getText()),
                                 strip_string_quotes(field.namedType().getText()))

        var = Variable(
            name='schema',
            value=schema_obj
        )
        self.codegen.write_variable(var)

    def __call__(self):
        try:
            self.codegen.import_package(package=graphene, mode=2, object='*')
            input_stream = FileStream(self.input_file)
            lexer = GraphQLLexer.GraphQLLexer(input_stream)
            stream = CommonTokenStream(lexer)
            parser = GraphQLParser(stream)
            tree = parser.document()
            walker = ParseTreeWalker()
            walker.walk(self, tree)
        except Exception as err:
            raise ParsingError(str(err))
